import logging
from collections.abc import Iterator, Sequence
from copy import deepcopy

from .typing.io import LLMPrompt
from .typing.message import Conversation, Message, SystemMessage

logger = logging.getLogger(__name__)


class MessageHistory:
    def __init__(self, sys_prompt: LLMPrompt | None = None) -> None:
        self._sys_prompt = sys_prompt
        self._batched_conversations: list[Conversation]
        self.reset()

    @property
    def sys_prompt(self) -> LLMPrompt | None:
        return self._sys_prompt

    def add_message_batch(self, message_batch: Sequence[Message]) -> None:
        """
        Adds a batch of messages to the current batched conversations.
        This method verifies that the size of the input message batch matches
        the expected batch size (self.batch_size).
        If there is a mismatch, the method adjusts by duplicating either
        the message or the conversation as necessary:

        - If the message batch contains exactly one message and
            self.batch_size > 1, the single message is duplicated to match
            the batch size.
        - If the message batch contains multiple messages but
            self.batch_size == 1, the entire conversation is duplicated to
            accommodate each message in the batch.
        - If the message batch size does not match self.batch_size and none of
            the above adjustments apply, a ValueError is raised.

        Afterwards, each message in the batch is appended to its corresponding
        conversation in the batched conversations.

        Args:
            message_batch: A sequence of Message objects
                representing the batch of messages to be added. Must align with
                or be adjusted to match the current batch size.

        Raises:
            ValueError: If the message batch size does not match the current
                batch size and cannot be automatically adjusted.

        """
        message_batch_size = len(message_batch)

        if message_batch_size == 1 and self.batch_size > 1:
            logger.info(
                "Message batch size is 1, current batch size is "
                f"{self.batch_size}: duplicating the message to match the "
                "current batch size"
            )
            message_batch = self._duplicate_message_to_current_batch_size(message_batch)
            message_batch_size = self.batch_size
        elif message_batch_size > 1 and self.batch_size == 1:
            logger.info(
                f"Message batch size is {len(message_batch)}, current batch "
                "size is 1: duplicating the conversation to match the message "
                "batch size"
            )
            self._duplicate_conversation_to_message_batch_size(message_batch_size)
        elif message_batch_size != self.batch_size:
            raise ValueError(
                f"Message batch size {message_batch_size} does not match "
                f"current batch size {self.batch_size}"
            )

        for batch_id in range(message_batch_size):
            self._batched_conversations[batch_id].append(message_batch[batch_id])

    def add_message_batches(self, message_batches: Sequence[Sequence[Message]]) -> None:
        for message_batch in message_batches:
            self.add_message_batch(message_batch)

    def add_message(self, message: Message) -> None:
        for conversation in self._batched_conversations:
            conversation.append(message)

    def add_messages(self, messages: Sequence[Message]) -> None:
        for message in messages:
            self.add_message(message)

    def __len__(self) -> int:
        return len(self._batched_conversations[0])

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(len={len(self)}; bs={self.batch_size})"

    def __getitem__(self, idx: int) -> tuple[Message, ...]:
        return tuple(conversation[idx] for conversation in self._batched_conversations)

    def __iter__(self) -> Iterator[tuple[Message, ...]]:
        for idx in range(len(self)):
            yield tuple(
                conversation[idx] for conversation in self._batched_conversations
            )

    def _duplicate_message_to_current_batch_size(
        self, message_batch: Sequence[Message]
    ) -> Sequence[Message]:
        assert len(message_batch) == 1, (
            "Message batch size must be 1 to duplicate to current batch size"
        )

        return [deepcopy(message_batch[0]) for _ in range(self.batch_size)]

    def _duplicate_conversation_to_message_batch_size(
        self, target_batch_size: int
    ) -> None:
        assert self.batch_size == 1, "Batch size must be 1 to duplicate conversation"
        self._batched_conversations = [
            deepcopy(self._batched_conversations[0]) for _ in range(target_batch_size)
        ]

    @property
    def batched_conversations(self) -> list[Conversation]:
        return self._batched_conversations

    @property
    def batch_size(self) -> int:
        return len(self._batched_conversations)

    def reset(
        self, sys_prompt: LLMPrompt | None = None, *, batch_size: int = 1
    ) -> None:
        if sys_prompt is not None:
            self._sys_prompt = sys_prompt

        conv: Conversation
        if self._sys_prompt is not None:
            conv = [SystemMessage(content=self._sys_prompt)]
        else:
            conv = []

        self._batched_conversations = [deepcopy(conv) for _ in range(batch_size)]

    def erase(self) -> None:
        self._batched_conversations = [[]]



from typing import TypeAlias

from ..typing.message import (
    AssistantMessage,
    SystemMessage,
    ToolMessage,
    Usage,
    UserMessage,
)
from ..typing.tool import ToolCall
from . import (
    ChatCompletionAssistantMessageParam,
    ChatCompletionDeveloperMessageParam,
    ChatCompletionFunctionMessageParam,
    ChatCompletionMessage,
    ChatCompletionMessageToolCallParam,
    ChatCompletionSystemMessageParam,
    ChatCompletionToolCallFunction,
    ChatCompletionToolMessageParam,
    ChatCompletionUsage,
    ChatCompletionUserMessageParam,
)
from .content_converters import from_api_content, to_api_content

OpenAIMessage: TypeAlias = (
    ChatCompletionAssistantMessageParam
    | ChatCompletionToolMessageParam
    | ChatCompletionUserMessageParam
    | ChatCompletionDeveloperMessageParam
    | ChatCompletionSystemMessageParam
    | ChatCompletionFunctionMessageParam
)


def from_api_user_message(
    api_message: ChatCompletionUserMessageParam, model_id: str | None = None
) -> UserMessage:
    content = from_api_content(api_message["content"])

    return UserMessage(content=content, model_id=model_id)


def to_api_user_message(message: UserMessage) -> ChatCompletionUserMessageParam:
    api_content = to_api_content(message.content)

    return ChatCompletionUserMessageParam(role="user", content=api_content)


def from_api_assistant_message(
    api_message: ChatCompletionMessage,
    api_usage: ChatCompletionUsage | None = None,
    model_id: str | None = None,
) -> AssistantMessage:
    usage = None
    if api_usage is not None:
        reasoning_tokens = None
        cached_tokens = None

        if api_usage.completion_tokens_details is not None:
            reasoning_tokens = api_usage.completion_tokens_details.reasoning_tokens
        if api_usage.prompt_tokens_details is not None:
            cached_tokens = api_usage.prompt_tokens_details.cached_tokens

        input_tokens = api_usage.prompt_tokens - (cached_tokens or 0)
        output_tokens = api_usage.completion_tokens - (reasoning_tokens or 0)

        usage = Usage(
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            reasoning_tokens=reasoning_tokens,
            cached_tokens=cached_tokens,
        )

    tool_calls = None
    if api_message.tool_calls is not None:
        tool_calls = [
            ToolCall(
                id=tool_call.id,
                tool_name=tool_call.function.name,
                tool_arguments=tool_call.function.arguments,
            )
            for tool_call in api_message.tool_calls
        ]

    return AssistantMessage(
        content=api_message.content,
        usage=usage,
        tool_calls=tool_calls,
        refusal=api_message.refusal,
        model_id=model_id,
    )


def to_api_assistant_message(
    message: AssistantMessage,
) -> ChatCompletionAssistantMessageParam:
    api_tool_calls = None
    if message.tool_calls is not None:
        api_tool_calls = [
            ChatCompletionMessageToolCallParam(
                type="function",
                id=tool_call.id,
                function=ChatCompletionToolCallFunction(
                    name=tool_call.tool_name,
                    arguments=tool_call.tool_arguments,
                ),
            )
            for tool_call in message.tool_calls
        ]

    api_message = ChatCompletionAssistantMessageParam(
        role="assistant",
        content=message.content,
        tool_calls=api_tool_calls or [],
        refusal=message.refusal,
    )
    if message.content is None and not api_tool_calls:
        # Some API providers return None in the generated content without errors,
        # even though None in the input content is not accepted.
        api_message["content"] = "<empty>"
    if api_tool_calls is None:
        api_message.pop("tool_calls")
    if message.refusal is None:
        api_message.pop("refusal")

    return api_message


def from_api_system_message(
    api_message: ChatCompletionSystemMessageParam,
    model_id: str | None = None,
) -> SystemMessage:
    return SystemMessage(content=api_message["content"], model_id=model_id)  # type: ignore


def to_api_system_message(
    message: SystemMessage,
) -> ChatCompletionSystemMessageParam:
    return ChatCompletionSystemMessageParam(role="system", content=message.content)


def from_api_tool_message(
    api_message: ChatCompletionToolMessageParam, model_id: str | None = None
) -> ToolMessage:
    return ToolMessage(
        content=api_message["content"],  # type: ignore
        tool_call_id=api_message["tool_call_id"],
        model_id=model_id,
    )


def to_api_tool_message(message: ToolMessage) -> ChatCompletionToolMessageParam:
    return ChatCompletionToolMessageParam(
        role="tool", content=message.content, tool_call_id=message.tool_call_id
    )



from collections.abc import Mapping
from typing import TypeAlias, TypeVar

from pydantic import BaseModel

from .content import ImageData

AgentID: TypeAlias = str


class AgentState(BaseModel):
    pass


class LLMPromptArgs(BaseModel):
    pass


InT = TypeVar("InT", contravariant=True)  # noqa: PLC0105
OutT = TypeVar("OutT", covariant=True)  # noqa: PLC0105
StateT = TypeVar("StateT", bound=AgentState, covariant=True)  # noqa: PLC0105

LLMPrompt: TypeAlias = str
LLMFormattedSystemArgs: TypeAlias = Mapping[str, str | int | bool]
LLMFormattedArgs: TypeAlias = Mapping[str, str | int | bool | ImageData]



import logging
from logging import Formatter, LogRecord
from pathlib import Path

import yaml
from termcolor import colored
from termcolor._types import Color  # type: ignore[import]


class ColorFormatter(Formatter):
    def format(self, record: LogRecord) -> str:
        message = super().format(record)
        color: Color | None = getattr(record, "color", None)
        if color:
            return colored(message, color)
        return message


def setup_logging(logs_file_path: str | Path, logs_config_path: str | Path) -> None:
    logs_file_path = Path(logs_file_path)
    logs_file_path.parent.mkdir(exist_ok=True, parents=True)
    with Path(logs_config_path).open() as f:
        config = yaml.safe_load(f)

    config["handlers"]["fileHandler"]["filename"] = logs_file_path

    logging.config.dictConfig(config)  # type: ignore

    root = logging.getLogger()
    for handler in root.handlers:
        if handler.formatter is not None:
            fmt_str = handler.formatter._fmt  # noqa: SLF001
            handler.setFormatter(ColorFormatter(fmt_str))


logger = logging.getLogger(__name__)



from typing import Any

from pydantic import BaseModel

from ..typing.tool import BaseTool, ToolChoice
from . import (
    ChatCompletionFunctionDefinition,
    ChatCompletionNamedToolChoiceFunction,
    ChatCompletionNamedToolChoiceParam,
    ChatCompletionToolChoiceOptionParam,
    ChatCompletionToolParam,
)


def to_api_tool(
    tool: BaseTool[BaseModel, Any, Any],
) -> ChatCompletionToolParam:
    function = ChatCompletionFunctionDefinition(
        name=tool.name,
        description=tool.description,
        parameters=tool.in_schema.model_json_schema(),
        strict=tool.strict,
    )
    if tool.strict is None:
        function.pop("strict")

    return ChatCompletionToolParam(type="function", function=function)


def to_api_tool_choice(
    tool_choice: ToolChoice,
) -> ChatCompletionToolChoiceOptionParam:
    if isinstance(tool_choice, BaseTool):
        return ChatCompletionNamedToolChoiceParam(
            type="function",
            function=ChatCompletionNamedToolChoiceFunction(name=tool_choice.name),
        )
    return tool_choice



# pyright: reportUnusedImport=false

from openai._streaming import (
    AsyncStream as ChatCompletionAsyncStream,  # type: ignore[import] # noqa: PLC2701
)
from openai.types import CompletionUsage as ChatCompletionUsage
from openai.types.chat.chat_completion import ChatCompletion
from openai.types.chat.chat_completion_assistant_message_param import (
    ChatCompletionAssistantMessageParam,
)
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk
from openai.types.chat.chat_completion_content_part_image_param import (
    ChatCompletionContentPartImageParam,
)
from openai.types.chat.chat_completion_content_part_image_param import (
    ImageURL as ChatCompletionImageURL,
)
from openai.types.chat.chat_completion_content_part_param import (
    ChatCompletionContentPartParam,
)
from openai.types.chat.chat_completion_content_part_text_param import (
    ChatCompletionContentPartTextParam,
)
from openai.types.chat.chat_completion_developer_message_param import (
    ChatCompletionDeveloperMessageParam,
)
from openai.types.chat.chat_completion_function_message_param import (
    ChatCompletionFunctionMessageParam,
)
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from openai.types.chat.chat_completion_message_param import (
    ChatCompletionMessageParam,
)
from openai.types.chat.chat_completion_message_tool_call_param import (
    ChatCompletionMessageToolCallParam,
)
from openai.types.chat.chat_completion_message_tool_call_param import (
    Function as ChatCompletionToolCallFunction,
)
from openai.types.chat.chat_completion_named_tool_choice_param import (
    ChatCompletionNamedToolChoiceParam,
)
from openai.types.chat.chat_completion_named_tool_choice_param import (
    Function as ChatCompletionNamedToolChoiceFunction,
)
from openai.types.chat.chat_completion_prediction_content_param import (
    ChatCompletionPredictionContentParam,
)
from openai.types.chat.chat_completion_stream_options_param import (
    ChatCompletionStreamOptionsParam,
)
from openai.types.chat.chat_completion_system_message_param import (
    ChatCompletionSystemMessageParam,
)
from openai.types.chat.chat_completion_tool_choice_option_param import (
    ChatCompletionToolChoiceOptionParam,
)
from openai.types.chat.chat_completion_tool_message_param import (
    ChatCompletionToolMessageParam,
)
from openai.types.chat.chat_completion_tool_param import (
    ChatCompletionToolParam,
)
from openai.types.chat.chat_completion_user_message_param import (
    ChatCompletionUserMessageParam,
)
from openai.types.chat.parsed_chat_completion import (
    ParsedChatCompletion,
    ParsedChatCompletionMessage,
    ParsedChoice,
)
from openai.types.shared_params.function_definition import (
    FunctionDefinition as ChatCompletionFunctionDefinition,
)
from openai.types.shared_params.response_format_json_object import (
    ResponseFormatJSONObject,
)
from openai.types.shared_params.response_format_json_schema import (
    ResponseFormatJSONSchema,
)
from openai.types.shared_params.response_format_text import (
    ResponseFormatText,
)

from .openai_llm import OpenAILLM, OpenAILLMSettings

__all__ = ["OpenAILLM", "OpenAILLMSettings"]



import asyncio
import functools
import logging
from collections.abc import Callable, Coroutine
from dataclasses import dataclass
from time import monotonic
from typing import Any, Generic, overload

from tqdm.autonotebook import tqdm

from ..utils import asyncio_gather_with_pbar
from .types import (
    P,
    ProcessorCallableList,
    ProcessorCallableSingle,
    R,
    RateLimWrapperWithArgsList,
    RateLimWrapperWithArgsSingle,
    T,
)
from .utils import partial_processor_callable, split_pos_args

logger = logging.getLogger(__name__)


@dataclass
class RateLimiterState:
    next_request_time: float = 0.0


class RateLimiterC(Generic[T, R]):
    def __init__(
        self,
        rpm: float,
        chunk_size: int = 1000,
        max_concurrency: int = 200,
    ):
        self._rpm = rpm
        self._max_concurrency = max_concurrency
        self._chunk_size = chunk_size

        self._lock = asyncio.Lock()
        self._state = RateLimiterState(next_request_time=0.0)
        self._semaphore = asyncio.Semaphore(self._max_concurrency)

    async def process_input(
        self,
        func_partial: Callable[[T], Coroutine[Any, Any, R]],
        inp: T,
    ) -> R:
        async with self._semaphore:
            async with self._lock:
                now = monotonic()
                if now < self._state.next_request_time:
                    await asyncio.sleep(self._state.next_request_time - now)
                self._state.next_request_time = monotonic() + 1.01 * 60.0 / self._rpm
            return await func_partial(inp)

    async def process_inputs(
        self,
        func_partial: Callable[[T], Coroutine[Any, Any, R]],
        inputs: list[T],
        no_tqdm: bool = False,
    ) -> list[R]:
        results: list[R] = []
        for i in tqdm(
            range(0, len(inputs), self._chunk_size),
            disable=no_tqdm,
            desc="Processing chunks",
        ):
            chunk = inputs[i : i + self._chunk_size]
            corouts = [
                self.process_input(func_partial=func_partial, inp=inp) for inp in chunk
            ]
            chunk_results = await asyncio.gather(*corouts)
            results.extend(chunk_results)

        return results

    @property
    def rpm(self) -> float:
        return self._rpm

    @rpm.setter
    def rpm(self, value: float) -> None:
        self._rpm = value

    @property
    def max_concurrency(self) -> int:
        return self._max_concurrency

    @property
    def chunk_size(self) -> int:
        return self._chunk_size

    @property
    def state(self) -> RateLimiterState:
        return self._state


@overload
def limit_rate(
    call: ProcessorCallableSingle[T, P, R],
    rate_limiter: RateLimiterC[T, R] | None = None,
) -> ProcessorCallableSingle[T, P, R]: ...


@overload
def limit_rate(
    call: None = None,
    rate_limiter: RateLimiterC[T, R] | None = None,
) -> RateLimWrapperWithArgsSingle[T, P, R]: ...


def limit_rate(
    call: ProcessorCallableSingle[T, P, R] | None = None,
    rate_limiter: RateLimiterC[T, R] | None = None,
) -> ProcessorCallableSingle[T, P, R] | RateLimWrapperWithArgsSingle[T, P, R]:
    if call is None:
        return functools.partial(limit_rate, rate_limiter=rate_limiter)

    @functools.wraps(call)  # type: ignore
    async def wrapper(*args: Any, **kwargs: Any) -> R:
        inp: T
        self_obj, inp, other_args = split_pos_args(call, args)
        call_partial = partial_processor_callable(call, self_obj, *other_args, **kwargs)

        _rate_limiter = rate_limiter
        if _rate_limiter is None:
            _rate_limiter = getattr(self_obj, "rate_limiter", None)

        if _rate_limiter is None:
            return await call_partial(inp)
        return await _rate_limiter.process_input(func_partial=call_partial, inp=inp)

    return wrapper


@overload
def limit_rate_chunked(
    call: ProcessorCallableList[T, P, R],
    rate_limiter: RateLimiterC[T, R] | None = None,
    no_tqdm: bool | None = None,
) -> ProcessorCallableList[T, P, R]: ...


@overload
def limit_rate_chunked(
    call: None = None,
    rate_limiter: RateLimiterC[T, R] | None = None,
    no_tqdm: bool | None = None,
) -> RateLimWrapperWithArgsList[T, P, R]: ...


def limit_rate_chunked(
    call: ProcessorCallableList[T, P, R] | None = None,
    rate_limiter: RateLimiterC[T, R] | None = None,
    no_tqdm: bool | None = None,
) -> ProcessorCallableList[T, P, R] | RateLimWrapperWithArgsList[T, P, R]:
    if call is None:
        return functools.partial(
            limit_rate_chunked, rate_limiter=rate_limiter, no_tqdm=no_tqdm
        )  # type: ignore

    @functools.wraps(call)  # type: ignore
    async def wrapper(*args: Any, **kwargs: Any) -> list[R]:
        assert call is not None

        self_obj, inputs, other_args = split_pos_args(call, args)
        call_partial = partial_processor_callable(call, self_obj, *other_args, **kwargs)

        _no_tqdm = no_tqdm
        _rate_limiter = rate_limiter
        if _no_tqdm is None:
            _no_tqdm = getattr(self_obj, "no_tqdm", False)
        if _rate_limiter is None:
            _rate_limiter = getattr(self_obj, "rate_limiter", None)

        if _rate_limiter is None:
            return await asyncio_gather_with_pbar(
                *[call_partial(inp) for inp in inputs], no_tqdm=_no_tqdm
            )
        return await _rate_limiter.process_inputs(
            func_partial=call_partial,  # type: ignore
            inputs=inputs,
            no_tqdm=_no_tqdm,
        )

    return wrapper



import ast
import asyncio
import json
import re
from collections.abc import Coroutine, Mapping
from datetime import datetime
from logging import getLogger
from pathlib import Path
from typing import Any, TypeVar

from pydantic import (
    GetCoreSchemaHandler,
    TypeAdapter,
    ValidationError,
)
from pydantic_core import core_schema
from tqdm.autonotebook import tqdm

logger = getLogger(__name__)

_JSON_START_RE = re.compile(r"[{\[]")

T = TypeVar("T")


def extract_json_substring(text: str) -> str | None:
    decoder = json.JSONDecoder()
    for match in _JSON_START_RE.finditer(text):
        start = match.start()
        try:
            _, end = decoder.raw_decode(text, idx=start)
            return text[start:end]
        except ValueError:
            continue

    return None


def parse_json_or_py_string(
    s: str, return_none_on_failure: bool = False
) -> dict[str, Any] | list[Any] | None:
    s_fmt = re.sub(r"```[a-zA-Z0-9]*\n|```", "", s).strip()
    try:
        return ast.literal_eval(s_fmt)
    except (ValueError, SyntaxError):
        try:
            return json.loads(s_fmt)
        except json.JSONDecodeError as exc:
            if return_none_on_failure:
                return None
            raise ValueError(
                "Invalid JSON/Python string - Both ast.literal_eval and json.loads "
                f"failed to parse the following response:\n{s}"
            ) from exc


def parse_json_or_py_substring(
    json_str: str, return_none_on_failure: bool = False
) -> dict[str, Any] | list[Any] | None:
    return parse_json_or_py_string(
        extract_json_substring(json_str) or "", return_none_on_failure
    )


def validate_obj_from_json_or_py_string(
    s: str, adapter: TypeAdapter[T], from_substring: bool = False
) -> T:
    try:
        if from_substring:
            parsed = parse_json_or_py_substring(s, return_none_on_failure=True)
        else:
            parsed = parse_json_or_py_string(s, return_none_on_failure=True)
        if parsed is None:
            parsed = s
        return adapter.validate_python(parsed)
    except (json.JSONDecodeError, ValidationError) as exc:
        raise ValueError(
            f"Invalid JSON or Python string:\n{s}\nExpected type: {adapter._type}",  # type: ignore[arg-type]
        ) from exc


def extract_xml_list(text: str) -> list[str]:
    pattern = re.compile(r"<(chunk_\d+)>(.*?)</\1>", re.DOTALL)

    chunks: list[str] = []
    for match in pattern.finditer(text):
        content = match.group(2).strip()
        chunks.append(content)
    return chunks


def build_marker_json_parser_type(
    marker_to_model: Mapping[str, type],
) -> type:
    """
    Return a Pydantic-compatible *type* that, when given a **str**, searches for
    the first marker substring and validates the JSON that follows with the
    corresponding Pydantic model.

    If no marker is found, the raw string is returned unchanged.

    Example:
    -------
    >>> Todo = build_marker_json_parser_type({'```json': MyModel})
    >>> Todo.validate('```json {"a": 1}')
    MyModel(a=1)

    """

    class MarkerParsedOutput:
        """String → (Model | str) parser generated by build_marker_json_parser_type."""

        @classmethod
        def __get_pydantic_core_schema__(
            cls,
            _source_type: Any,
            _handler: GetCoreSchemaHandler,
        ) -> core_schema.CoreSchema:
            def _validate(value: Any) -> Any:
                if not isinstance(value, str):
                    raise TypeError("MarkerParsedOutput expects a string")

                for marker, model in marker_to_model.items():
                    if marker in value:
                        adapter = TypeAdapter[Any](model)
                        return validate_obj_from_json_or_py_string(
                            value, adapter=adapter, from_substring=True
                        )

                return value

            return core_schema.no_info_after_validator_function(
                _validate, core_schema.any_schema()
            )

        @classmethod
        def __get_pydantic_json_schema__(
            cls,
            schema: core_schema.CoreSchema,
            handler: GetCoreSchemaHandler,
        ):
            return handler(schema)

    unique_suffix = "_".join(sorted(marker_to_model))[:40]
    MarkerParsedOutput.__name__ = f"MarkerParsedOutput_{unique_suffix}"

    return MarkerParsedOutput


def read_txt(file_path: str | Path, encoding: str = "utf-8") -> str:
    return Path(file_path).read_text(encoding=encoding)


def read_contents_from_file(
    file_path: str | Path,
    binary_mode: bool = False,
) -> str | bytes:
    try:
        if binary_mode:
            return Path(file_path).read_bytes()
        return Path(file_path).read_text()
    except FileNotFoundError:
        logger.error(f"File {file_path} not found.")
        return ""


def get_prompt(prompt_text: str | None, prompt_path: str | Path | None) -> str | None:
    if prompt_text is None:
        return read_contents_from_file(prompt_path) if prompt_path is not None else None  # type: ignore[arg-type]

    return prompt_text


async def asyncio_gather_with_pbar(
    *corouts: Coroutine[Any, Any, Any],
    no_tqdm: bool = False,
    desc: str | None = None,
) -> list[Any]:
    pbar = tqdm(total=len(corouts), desc=desc, disable=no_tqdm)

    async def run_and_update(coro: Coroutine[Any, Any, Any]) -> Any:
        result = await coro
        pbar.update(1)
        return result

    wrapped_tasks = [run_and_update(c) for c in corouts]
    results = await asyncio.gather(*wrapped_tasks)
    pbar.close()

    return results


def get_timestamp() -> str:
    return datetime.now().strftime("%Y%m%d_%H%M%S")



from collections.abc import AsyncIterator, Iterable
from typing import Any

from pydantic import BaseModel

from ..typing.completion import Completion, CompletionChunk
from ..typing.content import Content
from ..typing.converters import Converters
from ..typing.message import AssistantMessage, SystemMessage, ToolMessage, UserMessage
from ..typing.tool import BaseTool, ToolChoice
from . import (
    ChatCompletion,
    ChatCompletionAssistantMessageParam,
    ChatCompletionAsyncStream,  # type: ignore[import]
    ChatCompletionChunk,
    ChatCompletionContentPartParam,
    ChatCompletionMessage,
    ChatCompletionSystemMessageParam,
    ChatCompletionToolChoiceOptionParam,
    ChatCompletionToolMessageParam,
    ChatCompletionToolParam,
    ChatCompletionUsage,
    ChatCompletionUserMessageParam,
)
from .completion_converters import (
    from_api_completion,
    from_api_completion_chunk,
    from_api_completion_chunk_iterator,
    to_api_completion,
)
from .content_converters import from_api_content, to_api_content
from .message_converters import (
    from_api_assistant_message,
    from_api_system_message,
    from_api_tool_message,
    from_api_user_message,
    to_api_assistant_message,
    to_api_system_message,
    to_api_tool_message,
    to_api_user_message,
)
from .tool_converters import to_api_tool, to_api_tool_choice


class OpenAIConverters(Converters):
    @staticmethod
    def to_system_message(
        system_message: SystemMessage, **kwargs: Any
    ) -> ChatCompletionSystemMessageParam:
        return to_api_system_message(system_message, **kwargs)

    @staticmethod
    def from_system_message(
        raw_message: ChatCompletionSystemMessageParam,
        model_id: str | None = None,
        **kwargs: Any,
    ) -> SystemMessage:
        return from_api_system_message(raw_message, model_id=model_id, **kwargs)

    @staticmethod
    def to_user_message(
        user_message: UserMessage, **kwargs: Any
    ) -> ChatCompletionUserMessageParam:
        return to_api_user_message(user_message, **kwargs)

    @staticmethod
    def from_user_message(
        raw_message: ChatCompletionUserMessageParam,
        model_id: str | None = None,
        **kwargs: Any,
    ) -> UserMessage:
        return from_api_user_message(raw_message, model_id=model_id, **kwargs)

    @staticmethod
    def to_assistant_message(
        assistant_message: AssistantMessage, **kwargs: Any
    ) -> ChatCompletionAssistantMessageParam:
        return to_api_assistant_message(assistant_message, **kwargs)

    @staticmethod
    def from_assistant_message(
        raw_message: ChatCompletionMessage,
        raw_usage: ChatCompletionUsage,
        model_id: str | None = None,
        **kwargs: Any,
    ) -> AssistantMessage:
        return from_api_assistant_message(
            raw_message, raw_usage, model_id=model_id, **kwargs
        )

    @staticmethod
    def to_tool_message(
        tool_message: ToolMessage, **kwargs: Any
    ) -> ChatCompletionToolMessageParam:
        return to_api_tool_message(tool_message, **kwargs)

    @staticmethod
    def from_tool_message(
        raw_message: ChatCompletionToolMessageParam,
        model_id: str | None = None,
        **kwargs: Any,
    ) -> ToolMessage:
        return from_api_tool_message(raw_message, model_id=model_id, **kwargs)

    @staticmethod
    def to_tool(
        tool: BaseTool[BaseModel, Any, Any], **kwargs: Any
    ) -> ChatCompletionToolParam:
        return to_api_tool(tool, **kwargs)

    @staticmethod
    def to_tool_choice(
        tool_choice: ToolChoice, **kwargs: Any
    ) -> ChatCompletionToolChoiceOptionParam:
        return to_api_tool_choice(tool_choice, **kwargs)

    @staticmethod
    def to_content(
        content: Content, **kwargs: Any
    ) -> Iterable[ChatCompletionContentPartParam]:
        return to_api_content(content, **kwargs)

    @staticmethod
    def from_content(
        raw_content: str | Iterable[ChatCompletionContentPartParam],
        **kwargs: Any,
    ) -> Content:
        return from_api_content(raw_content, **kwargs)

    @staticmethod
    def to_completion(completion: Completion, **kwargs: Any) -> ChatCompletion:
        return to_api_completion(completion, **kwargs)

    @staticmethod
    def from_completion(
        raw_completion: ChatCompletion,
        model_id: str | None = None,
        **kwargs: Any,
    ) -> Completion:
        return from_api_completion(raw_completion, model_id=model_id, **kwargs)

    @staticmethod
    def to_completion_chunk(
        chunk: CompletionChunk, **kwargs: Any
    ) -> ChatCompletionChunk:
        raise NotImplementedError

    @staticmethod
    def from_completion_chunk(
        raw_chunk: ChatCompletionChunk,
        model_id: str | None = None,
        **kwargs: Any,
    ) -> CompletionChunk:
        return from_api_completion_chunk(raw_chunk, model_id=model_id, **kwargs)

    @staticmethod
    def from_completion_chunk_iterator(  # type: ignore[override]
        raw_chunk_iterator: ChatCompletionAsyncStream[ChatCompletionChunk],
        model_id: str | None = None,
        **kwargs: Any,
    ) -> AsyncIterator[CompletionChunk]:
        return from_api_completion_chunk_iterator(
            raw_chunk_iterator, model_id=model_id, **kwargs
        )



from collections.abc import Sequence
from typing import Any, Generic, TypeAlias, TypeVar
from uuid import uuid4

from pydantic import BaseModel, ConfigDict, Field, PrivateAttr

from .printer import Printer
from .typing.content import ImageData
from .typing.io import (
    AgentID,
    AgentState,
    InT,
    LLMPrompt,
    LLMPromptArgs,
    OutT,
    StateT,
)
from .usage_tracker import UsageTracker

SystemRunArgs: TypeAlias = LLMPromptArgs
UserRunArgs: TypeAlias = LLMPromptArgs | list[LLMPromptArgs]


class RunArgs(BaseModel):
    sys: SystemRunArgs = Field(default_factory=LLMPromptArgs)
    usr: UserRunArgs = Field(default_factory=LLMPromptArgs)

    model_config = ConfigDict(extra="forbid")


class InteractionRecord(BaseModel, Generic[InT, OutT, StateT]):
    source_id: str
    recipient_ids: Sequence[AgentID]
    state: StateT
    chat_inputs: LLMPrompt | Sequence[str | ImageData] | None = None
    sys_prompt: LLMPrompt | None = None
    in_prompt: LLMPrompt | None = None
    sys_args: SystemRunArgs | None = None
    usr_args: UserRunArgs | None = None
    in_args: Sequence[InT] | None = None
    outputs: Sequence[OutT]

    model_config = ConfigDict(extra="forbid", frozen=True)


InteractionHistory: TypeAlias = list[InteractionRecord[Any, Any, AgentState]]


CtxT = TypeVar("CtxT")


class RunContextWrapper(BaseModel, Generic[CtxT]):
    context: CtxT | None = None
    run_id: str = Field(default_factory=lambda: str(uuid4())[:8], frozen=True)
    run_args: dict[AgentID, RunArgs] = Field(default_factory=dict)
    interaction_history: InteractionHistory = Field(default_factory=list)  # type: ignore[valid-type]

    print_messages: bool = False

    _usage_tracker: UsageTracker = PrivateAttr()
    _printer: Printer = PrivateAttr()

    def model_post_init(self, context: Any) -> None:  # noqa: ARG002
        self._usage_tracker = UsageTracker(source_id=self.run_id)
        self._printer = Printer(
            source_id=self.run_id, print_messages=self.print_messages
        )

    @property
    def usage_tracker(self) -> UsageTracker:
        return self._usage_tracker

    @property
    def printer(self) -> Printer:
        return self._printer

    model_config = ConfigDict(extra="forbid")



from abc import ABC, abstractmethod
from collections.abc import AsyncIterator
from typing import Any

from pydantic import BaseModel

from .completion import Completion, CompletionChunk
from .content import Content
from .message import AssistantMessage, Message, SystemMessage, ToolMessage, UserMessage
from .tool import BaseTool, ToolChoice


class Converters(ABC):
    @staticmethod
    @abstractmethod
    def to_system_message(system_message: SystemMessage, **kwargs: Any) -> Any:
        pass

    @staticmethod
    @abstractmethod
    def from_system_message(raw_message: Any, **kwargs: Any) -> SystemMessage:
        pass

    @staticmethod
    @abstractmethod
    def to_user_message(user_message: UserMessage, **kwargs: Any) -> Any:
        pass

    @staticmethod
    @abstractmethod
    def from_user_message(raw_message: Any, **kwargs: Any) -> UserMessage:
        pass

    @staticmethod
    @abstractmethod
    def to_assistant_message(assistant_message: AssistantMessage, **kwargs: Any) -> Any:
        pass

    @staticmethod
    @abstractmethod
    def from_assistant_message(
        raw_message: Any, raw_usage: Any, **kwargs: Any
    ) -> AssistantMessage:
        pass

    @staticmethod
    @abstractmethod
    def to_tool_message(tool_message: ToolMessage, **kwargs: Any) -> Any:
        pass

    @staticmethod
    @abstractmethod
    def from_tool_message(raw_message: Any, **kwargs: Any) -> ToolMessage:
        pass

    @classmethod
    def to_message(cls, message: Message, **kwargs: Any) -> Any:
        if isinstance(message, UserMessage):
            return cls.to_user_message(message, **kwargs)
        if isinstance(message, AssistantMessage):
            return cls.to_assistant_message(message, **kwargs)
        if isinstance(message, ToolMessage):
            return cls.to_tool_message(message, **kwargs)

        return cls.to_system_message(message, **kwargs)

    @staticmethod
    @abstractmethod
    def to_tool(tool: BaseTool[BaseModel, Any, Any], **kwargs: Any) -> Any:
        pass

    @staticmethod
    @abstractmethod
    def to_tool_choice(tool_choice: ToolChoice, **kwargs: Any) -> Any:
        pass

    @staticmethod
    @abstractmethod
    def to_content(content: Content, **kwargs: Any) -> Any:
        pass

    @staticmethod
    @abstractmethod
    def from_content(raw_content: Any, **kwargs: Any) -> Content:
        pass

    @staticmethod
    @abstractmethod
    def to_completion(completion: Completion, **kwargs: Any) -> Any:
        pass

    @staticmethod
    @abstractmethod
    def from_completion(raw_completion: Any, **kwargs: Any) -> Completion:
        pass

    @staticmethod
    @abstractmethod
    def to_completion_chunk(chunk: CompletionChunk, **kwargs: Any) -> Any:
        pass

    @staticmethod
    @abstractmethod
    def from_completion_chunk(raw_chunk: Any, **kwargs: Any) -> CompletionChunk:
        pass

    @staticmethod
    @abstractmethod
    def from_completion_chunk_iterator(
        raw_chunk_iterator: AsyncIterator[Any], **kwargs: Any
    ) -> AsyncIterator[CompletionChunk]:
        pass



from collections.abc import Callable, Coroutine
from typing import Any, Concatenate, ParamSpec, TypeAlias, TypeVar

T = TypeVar("T")
R = TypeVar("R")
P = ParamSpec("P")

ProcessorFuncSingle: TypeAlias = Callable[Concatenate[T, P], Coroutine[Any, Any, R]]
ProcessorFuncList: TypeAlias = Callable[
    Concatenate[list[T], P], Coroutine[Any, Any, list[R]]
]

ProcessorMethodSingle: TypeAlias = Callable[
    Concatenate[Any, T, P], Coroutine[Any, Any, R]
]
ProcessorMethodList: TypeAlias = Callable[
    Concatenate[Any, list[T], P], Coroutine[Any, Any, list[R]]
]

ProcessorCallableSingle: TypeAlias = (
    ProcessorFuncSingle[T, P, R] | ProcessorMethodSingle[T, P, R]
)

ProcessorCallableList: TypeAlias = (
    ProcessorFuncList[T, P, R] | ProcessorMethodList[T, P, R]
)


RateLimWrapperWithArgsSingle = Callable[
    [ProcessorCallableSingle[T, P, R]], ProcessorCallableSingle[T, P, R]
]


RateLimWrapperWithArgsList = Callable[
    [ProcessorCallableList[T, P, R]], ProcessorCallableList[T, P, R]
]



import fnmatch
import logging
import os
from abc import abstractmethod
from collections.abc import AsyncIterator, Sequence
from copy import deepcopy
from typing import Any, Generic, Literal

import httpx
from pydantic import BaseModel, TypeAdapter
from tenacity import (
    RetryCallState,
    retry,
    stop_after_attempt,
    wait_random_exponential,
)
from typing_extensions import TypedDict

from .http_client import AsyncHTTPClientParams, create_async_http_client
from .llm import LLM, ConvertT, LLMSettings, SettingsT
from .memory import MessageHistory
from .rate_limiting.rate_limiter_chunked import (  # type: ignore
    RateLimiterC,
    limit_rate_chunked,
)
from .typing.completion import Completion, CompletionChunk
from .typing.message import AssistantMessage, Conversation
from .typing.tool import BaseTool, ToolChoice
from .utils import validate_obj_from_json_or_py_string

logger = logging.getLogger(__name__)


APIProvider = Literal["openai", "openrouter", "google_ai_studio"]


class APIProviderInfo(TypedDict):
    name: APIProvider
    base_url: str
    api_key: str | None
    struct_output_support: tuple[str, ...]


PROVIDERS: dict[APIProvider, APIProviderInfo] = {
    "openai": APIProviderInfo(
        name="openai",
        base_url="https://api.openai.com/v1",
        api_key=os.getenv("OPENAI_API_KEY"),
        struct_output_support=("*",),
    ),
    "openrouter": APIProviderInfo(
        name="openrouter",
        base_url="https://openrouter.ai/api/v1",
        api_key=os.getenv("OPENROUTER_API_KEY"),
        struct_output_support=(),
    ),
    "google_ai_studio": APIProviderInfo(
        name="google_ai_studio",
        base_url="https://generativelanguage.googleapis.com/v1beta/openai/",
        api_key=os.getenv("GOOGLE_AI_STUDIO_API_KEY"),
        struct_output_support=("*",),
    ),
}


def retry_error_callback(retry_state: RetryCallState) -> None:
    assert retry_state.outcome is not None
    exception = retry_state.outcome.exception()
    if exception:
        if retry_state.attempt_number == 1:
            logger.error(
                f"CloudLLM completion request failed:\n{exception}",
                exc_info=exception,
            )
        if retry_state.attempt_number > 1:
            logger.warning(
                f"CloudLLM completion request failed after retrying:\n{exception}",
                exc_info=exception,
            )


def retry_before_callback(retry_state: RetryCallState) -> None:
    if retry_state.attempt_number > 1:
        logger.info(
            "Retrying CloudLLM completion request "
            f"(attempt {retry_state.attempt_number - 1}) ..."
        )


class CloudLLMSettings(LLMSettings, total=False):
    max_completion_tokens: int | None
    temperature: float | None
    top_p: float | None
    seed: int | None
    use_structured_outputs: bool


class CloudLLM(LLM[SettingsT, ConvertT], Generic[SettingsT, ConvertT]):
    def __init__(
        self,
        # Base LLM args
        model_name: str,
        converters: ConvertT,
        llm_settings: SettingsT | None = None,
        model_id: str | None = None,
        tools: list[BaseTool[BaseModel, Any, Any]] | None = None,
        response_format: type | None = None,
        # Connection settings
        async_http_client_params: (
            dict[str, Any] | AsyncHTTPClientParams | None
        ) = None,
        # Rate limiting
        rate_limiter: (RateLimiterC[Conversation, AssistantMessage] | None) = None,
        rate_limiter_rpm: float | None = None,
        rate_limiter_chunk_size: int = 1000,
        rate_limiter_max_concurrency: int = 300,
        # Retries
        num_generation_retries: int = 0,
        # Disable tqdm for batch processing
        no_tqdm: bool = True,
        **kwargs: Any,
    ) -> None:
        self.llm_settings: CloudLLMSettings | None

        super().__init__(
            model_name=model_name,
            llm_settings=llm_settings,
            converters=converters,
            model_id=model_id,
            tools=tools,
            response_format=response_format,
            **kwargs,
        )

        self._model_name = model_name

        api_provider = model_name.split(":", 1)[0]
        api_model_name = model_name.split(":", 1)[-1]
        if api_provider not in PROVIDERS:
            raise ValueError(
                f"API provider '{api_provider}' is not supported. "
                f"Supported providers are: {', '.join(PROVIDERS.keys())}"
            )
        self._api_provider: APIProvider = api_provider
        self._api_model_name: str = api_model_name

        self._struct_output_support: bool = any(
            fnmatch.fnmatch(self._model_name, pat)
            for pat in PROVIDERS[api_provider]["struct_output_support"]
        )
        self._response_format_pyd: TypeAdapter[Any] | None = (
            TypeAdapter(self._response_format) if response_format else None
        )
        if (
            self._llm_settings.get("use_structured_outputs")
            and not self._struct_output_support
        ):
            raise ValueError(
                f"Model {api_provider}:{self._model_name} does "
                "not support structured outputs."
            )

        self._rate_limiter: RateLimiterC[Conversation, AssistantMessage] | None = (
            self._get_rate_limiter(
                rate_limiter=rate_limiter,
                rate_limiter_rpm=rate_limiter_rpm,
                rate_limiter_chunk_size=rate_limiter_chunk_size,
                rate_limiter_max_concurrency=rate_limiter_max_concurrency,
            )
        )
        self.no_tqdm = no_tqdm

        self._base_url: str = PROVIDERS[api_provider]["base_url"]
        self._api_key: str | None = PROVIDERS[api_provider]["api_key"]
        self._client: Any

        self._async_http_client: httpx.AsyncClient | None = None
        if async_http_client_params is not None:
            val_async_http_client_params = AsyncHTTPClientParams.model_validate(
                async_http_client_params
            )
            self._async_http_client = create_async_http_client(
                val_async_http_client_params
            )

        self.num_generation_retries = num_generation_retries

    @property
    def api_provider(self) -> APIProvider:
        return self._api_provider

    @property
    def rate_limiter(
        self,
    ) -> RateLimiterC[Conversation, AssistantMessage] | None:
        return self._rate_limiter

    def _make_completion_kwargs(
        self, conversation: Conversation, tool_choice: ToolChoice | None = None
    ) -> dict[str, Any]:
        api_messages = [self._converters.to_message(m) for m in conversation]

        api_tools = None
        api_tool_choice = None
        if self.tools:
            api_tools = [self._converters.to_tool(t) for t in self.tools.values()]
            if tool_choice is not None:
                api_tool_choice = self._converters.to_tool_choice(tool_choice)

        api_llm_settings = deepcopy(self.llm_settings or {})
        api_llm_settings.pop("use_structured_outputs", None)

        return dict(
            api_messages=api_messages,
            api_tools=api_tools,
            api_tool_choice=api_tool_choice,
            api_response_format=self._response_format,
            **api_llm_settings,
        )

    @abstractmethod
    async def _get_completion(
        self,
        api_messages: list[Any],
        *,
        api_tools: list[Any] | None = None,
        api_tool_choice: Any | None = None,
        **api_llm_settings: Any,
    ) -> Any:
        pass

    @abstractmethod
    async def _get_parsed_completion(
        self,
        api_messages: list[Any],
        *,
        api_tools: list[Any] | None = None,
        api_tool_choice: Any | None = None,
        api_response_format: type | None = None,
        **api_llm_settings: Any,
    ) -> Any:
        pass

    @abstractmethod
    async def _get_completion_stream(
        self,
        api_messages: list[Any],
        *,
        api_tools: list[Any] | None = None,
        api_tool_choice: Any | None = None,
        **api_llm_settings: Any,
    ) -> AsyncIterator[Any]:
        pass

    async def generate_completion(
        self,
        conversation: Conversation,
        *,
        tool_choice: ToolChoice | None = None,
        **kwargs: Any,
    ) -> Completion:
        completion_kwargs = self._make_completion_kwargs(
            conversation=conversation, tool_choice=tool_choice
        )

        if (
            self._response_format is None
            or (not self._struct_output_support)
            or (not self._llm_settings.get("use_structured_outputs"))
        ):
            completion_kwargs.pop("api_response_format", None)
            api_completion = await self._get_completion(**completion_kwargs, **kwargs)
        else:
            api_completion = await self._get_parsed_completion(
                **completion_kwargs, **kwargs
            )

        completion = self._converters.from_completion(
            api_completion, model_id=self.model_id
        )

        self._validate_completion(completion)

        return completion

    def _validate_completion(self, completion: Completion) -> None:
        for choice in completion.choices:
            message = choice.message
            if (
                self._response_format_pyd is not None
                and not self._llm_settings.get("use_structured_outputs")
                and not message.tool_calls
            ):
                validate_obj_from_json_or_py_string(
                    message.content or "",
                    adapter=self._response_format_pyd,
                    from_substring=True,
                )

    async def generate_completion_stream(
        self,
        conversation: Conversation,
        *,
        tool_choice: ToolChoice | None = None,
        **kwargs: Any,
    ) -> AsyncIterator[CompletionChunk]:
        completion_kwargs = self._make_completion_kwargs(
            conversation=conversation, tool_choice=tool_choice
        )
        completion_kwargs.pop("api_response_format", None)
        api_completion_chunk_iterator = await self._get_completion_stream(
            **completion_kwargs, **kwargs
        )

        return self._converters.from_completion_chunk_iterator(
            api_completion_chunk_iterator, model_id=self.model_id
        )

    async def _generate_completion_with_retry(
        self,
        conversation: Conversation,
        *,
        tool_choice: ToolChoice | None = None,
        **kwargs: Any,
    ) -> Completion:
        wrapped_func = retry(
            wait=wait_random_exponential(min=1, max=8),
            stop=stop_after_attempt(self.num_generation_retries + 1),
            before=retry_before_callback,
            retry_error_callback=retry_error_callback,
        )(self.__class__.generate_completion)

        return await wrapped_func(self, conversation, tool_choice=tool_choice, **kwargs)

    @limit_rate_chunked  # type: ignore
    async def _generate_completion_batch_with_retry_and_rate_lim(
        self,
        conversation: Conversation,
        *,
        tool_choice: ToolChoice | None = None,
        **kwargs: Any,
    ) -> Completion:
        return await self._generate_completion_with_retry(
            conversation, tool_choice=tool_choice, **kwargs
        )

    async def generate_completion_batch(
        self,
        message_history: MessageHistory,
        *,
        tool_choice: ToolChoice | None = None,
        **kwargs: Any,
    ) -> Sequence[Completion]:
        return await self._generate_completion_batch_with_retry_and_rate_lim(
            list(message_history.batched_conversations),  # type: ignore
            tool_choice=tool_choice,
            **kwargs,
        )

    async def generate_message(
        self,
        conversation: Conversation,
        *,
        tool_choice: ToolChoice | None = None,
        **kwargs: Any,
    ) -> AssistantMessage:
        completion = await self.generate_completion(
            conversation, tool_choice=tool_choice, **kwargs
        )

        return completion.choices[0].message

    async def generate_message_batch(
        self,
        message_history: MessageHistory,
        *,
        tool_choice: ToolChoice | None = None,
        **kwargs: Any,
    ) -> Sequence[AssistantMessage]:
        completion_batch = await self.generate_completion_batch(
            message_history, tool_choice=tool_choice, **kwargs
        )

        return [completion.choices[0].message for completion in completion_batch]

    def _get_rate_limiter(
        self,
        rate_limiter: RateLimiterC[Conversation, AssistantMessage] | None = None,
        rate_limiter_rpm: float | None = None,
        rate_limiter_chunk_size: int = 1000,
        rate_limiter_max_concurrency: int = 300,
    ) -> RateLimiterC[Conversation, AssistantMessage] | None:
        if rate_limiter is not None:
            logger.info(
                f"[{self.__class__.__name__}] Set rate limit to {rate_limiter.rpm} RPM"
            )
            return rate_limiter
        if rate_limiter_rpm is not None:
            logger.info(
                f"[{self.__class__.__name__}] Set rate limit to {rate_limiter_rpm} RPM"
            )
            return RateLimiterC(
                rpm=rate_limiter_rpm,
                chunk_size=rate_limiter_chunk_size,
                max_concurrency=rate_limiter_max_concurrency,
            )

        return None



from abc import ABC, abstractmethod
from typing import Any, ClassVar, Generic

from pydantic import TypeAdapter

from .generics_utils import AutoInstanceAttributesMixin
from .run_context import CtxT, RunContextWrapper
from .typing.io import AgentID, OutT, StateT
from .typing.tool import BaseTool


class BaseAgent(AutoInstanceAttributesMixin, ABC, Generic[OutT, StateT, CtxT]):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {0: "_out_type"}

    @abstractmethod
    def __init__(self, agent_id: AgentID, **kwargs: Any) -> None:
        self._out_type: type[OutT]
        self._state: StateT

        super().__init__()

        self._agent_id = agent_id
        self._out_type_adapter: TypeAdapter[OutT] = TypeAdapter(self._out_type)

    @property
    def out_type(self) -> type[OutT]:
        return self._out_type

    @property
    def agent_id(self) -> AgentID:
        return self._agent_id

    @property
    def state(self) -> StateT:
        return self._state

    @abstractmethod
    async def run(
        self,
        chat_inputs: Any | None = None,
        *,
        ctx: RunContextWrapper[CtxT] | None = None,
        **kwargs: Any,
    ) -> Any:
        pass

    @abstractmethod
    def as_tool(
        self, tool_name: str, tool_description: str, tool_strict: bool = True
    ) -> BaseTool[Any, OutT, CtxT]:
        pass



import asyncio
import json
from collections.abc import Coroutine, Sequence
from logging import getLogger
from typing import Any, Generic, Protocol

from pydantic import BaseModel

from .llm import LLM, LLMSettings
from .llm_agent_state import LLMAgentState
from .run_context import CtxT, RunContextWrapper
from .typing.converters import Converters
from .typing.message import AssistantMessage, Conversation, Message, ToolMessage
from .typing.tool import BaseTool, ToolCall, ToolChoice

logger = getLogger(__name__)


class ExitToolCallLoopHandler(Protocol[CtxT]):
    def __call__(
        self,
        conversation: Conversation,
        *,
        ctx: RunContextWrapper[CtxT] | None,
        **kwargs: Any,
    ) -> bool: ...


class ManageAgentStateHandler(Protocol[CtxT]):
    def __call__(
        self,
        state: LLMAgentState,
        *,
        ctx: RunContextWrapper[CtxT] | None,
        **kwargs: Any,
    ) -> None: ...


class ToolOrchestrator(Generic[CtxT]):
    def __init__(
        self,
        agent_id: str,
        llm: LLM[LLMSettings, Converters],
        tools: list[BaseTool[BaseModel, Any, CtxT]] | None,
        max_turns: int,
        react_mode: bool = False,
    ) -> None:
        self._agent_id = agent_id

        self._llm = llm
        self._llm.tools = tools

        self._max_turns = max_turns
        self._react_mode = react_mode

        self.exit_tool_call_loop_impl: ExitToolCallLoopHandler[CtxT] | None = None
        self.manage_agent_state_impl: ManageAgentStateHandler[CtxT] | None = None

    @property
    def agent_id(self) -> str:
        return self._agent_id

    @property
    def llm(self) -> LLM[LLMSettings, Converters]:
        return self._llm

    @property
    def tools(self) -> dict[str, BaseTool[BaseModel, Any, CtxT]]:
        return self._llm.tools or {}

    @property
    def max_turns(self) -> int:
        return self._max_turns

    def _exit_tool_call_loop(
        self,
        conversation: Conversation,
        *,
        ctx: RunContextWrapper[CtxT] | None = None,
        **kwargs: Any,
    ) -> bool:
        if self.exit_tool_call_loop_impl:
            return self.exit_tool_call_loop_impl(
                conversation=conversation, ctx=ctx, **kwargs
            )

        assert conversation, "Conversation must not be empty"
        assert isinstance(conversation[-1], AssistantMessage), (
            "Last message in conversation must be an AssistantMessage"
        )

        return not bool(conversation[-1].tool_calls)

    def _manage_agent_state(
        self,
        state: LLMAgentState,
        *,
        ctx: RunContextWrapper[CtxT] | None = None,
        **kwargs: Any,
    ) -> None:
        if self.manage_agent_state_impl:
            self.manage_agent_state_impl(state=state, ctx=ctx, **kwargs)

    async def generate_once(
        self,
        state: LLMAgentState,
        tool_choice: ToolChoice | None = None,
        ctx: RunContextWrapper[CtxT] | None = None,
    ) -> Sequence[AssistantMessage]:
        message_history = state.message_history
        message_batch = await self.llm.generate_message_batch(
            message_history, tool_choice=tool_choice
        )
        message_history.add_message_batch(message_batch)

        self._print_messages_and_track_usage(message_batch, ctx=ctx)

        return message_batch

    async def run_loop(
        self,
        state: LLMAgentState,
        ctx: RunContextWrapper[CtxT] | None = None,
    ) -> None:
        message_history = state.message_history
        assert message_history.batch_size == 1, (
            "Batch size must be 1 for tool call loop"
        )

        tool_choice: ToolChoice

        tool_choice = "none" if self._react_mode else "auto"
        gen_message_batch = await self.generate_once(
            state, tool_choice=tool_choice, ctx=ctx
        )

        turns = 0

        while True:
            self._manage_agent_state(state=state, ctx=ctx, num_turns=turns)

            if self._exit_tool_call_loop(
                message_history.batched_conversations[0], ctx=ctx, num_turns=turns
            ):
                return
            if turns >= self.max_turns:
                logger.info(
                    f"Max turns reached: {self.max_turns}. Stopping tool call loop."
                )
                return

            msg = gen_message_batch[0]
            if msg.tool_calls:
                tool_messages = await self.call_tools(msg.tool_calls, ctx=ctx)
                message_history.add_messages(tool_messages)

            tool_choice = "none" if (self._react_mode and msg.tool_calls) else "auto"
            gen_message_batch = await self.generate_once(
                state, tool_choice=tool_choice, ctx=ctx
            )

            turns += 1

    async def call_tools(
        self,
        calls: Sequence[ToolCall],
        ctx: RunContextWrapper[CtxT] | None = None,
    ) -> Sequence[ToolMessage]:
        corouts: list[Coroutine[Any, Any, BaseModel]] = []
        for call in calls:
            tool = self.tools[call.tool_name]
            args = json.loads(call.tool_arguments)
            corouts.append(tool(ctx=ctx, **args))

        outs = await asyncio.gather(*corouts)

        tool_messages = [
            ToolMessage.from_tool_output(out, call, model_id=self.agent_id)
            for out, call in zip(outs, calls, strict=False)
        ]

        self._print_messages(tool_messages, ctx=ctx)

        return tool_messages

    def _print_messages(
        self,
        message_batch: Sequence[Message],
        ctx: RunContextWrapper[CtxT] | None = None,
    ) -> None:
        if ctx:
            ctx.printer.print_llm_messages(message_batch, agent_id=self.agent_id)

    def _print_messages_and_track_usage(
        self,
        message_batch: Sequence[AssistantMessage],
        ctx: RunContextWrapper[CtxT] | None = None,
    ) -> None:
        if ctx:
            self._print_messages(message_batch, ctx=ctx)
            ctx.usage_tracker.update(
                messages=message_batch, model_name=self.llm.model_name
            )



from collections.abc import AsyncIterator

from ..typing.completion import Completion, CompletionChoice, CompletionChunk
from . import (
    ChatCompletion,
    ChatCompletionAsyncStream,  # type: ignore[import]
    ChatCompletionChunk,
)
from .message_converters import from_api_assistant_message


def from_api_completion(
    api_completion: ChatCompletion, model_id: str | None = None
) -> Completion:
    choices: list[CompletionChoice] = []
    if api_completion.choices is None:  # type: ignore
        # Some providers return None for the choices when there is an error
        # TODO: add custom error types
        raise RuntimeError(
            f"Completion API error: {getattr(api_completion, 'error', None)}"
        )
    for api_choice in api_completion.choices:
        # TODO: currently no way to assign individual message usages when len(choices) > 1
        finish_reason = api_choice.finish_reason
        # Some providers return None for the message when finish_reason is other than "stop"
        if api_choice.message is None:  # type: ignore
            raise RuntimeError(
                f"API returned None for message with finish_reason: {finish_reason}"
            )
        message = from_api_assistant_message(
            api_choice.message, api_completion.usage, model_id=model_id
        )
        choices.append(CompletionChoice(message=message, finish_reason=finish_reason))

    return Completion(choices=choices, model_id=model_id)


def to_api_completion(completion: Completion) -> ChatCompletion:
    raise NotImplementedError


def from_api_completion_chunk(
    api_completion_chunk: ChatCompletionChunk, model_id: str | None = None
) -> CompletionChunk:
    delta = api_completion_chunk.choices[0].delta.content

    return CompletionChunk(delta=delta, model_id=model_id)


async def from_api_completion_chunk_iterator(
    api_completion_chunk_iterator: ChatCompletionAsyncStream[ChatCompletionChunk],
    model_id: str | None = None,
) -> AsyncIterator[CompletionChunk]:
    async for api_chunk in api_completion_chunk_iterator:
        yield from_api_completion_chunk(api_chunk, model_id=model_id)



import json
from collections.abc import Hashable, Mapping, Sequence
from enum import StrEnum
from typing import Annotated, Any, Literal, TypeAlias
from uuid import uuid4

from pydantic import BaseModel, Field, NonNegativeFloat, NonNegativeInt
from pydantic.json import pydantic_encoder

from .content import Content, ImageData
from .tool import ToolCall


class Role(StrEnum):
    USER = "user"
    SYSTEM = "system"
    ASSISTANT = "assistant"
    TOOL = "tool"


class Usage(BaseModel):
    input_tokens: NonNegativeInt = 0
    output_tokens: NonNegativeInt = 0
    reasoning_tokens: NonNegativeInt | None = None
    cached_tokens: NonNegativeInt | None = None
    cost: NonNegativeFloat | None = None

    def __add__(self, add_usage: "Usage") -> "Usage":
        input_tokens = self.input_tokens + add_usage.input_tokens
        output_tokens = self.output_tokens + add_usage.output_tokens
        if self.reasoning_tokens is not None or add_usage.reasoning_tokens is not None:
            reasoning_tokens = (self.reasoning_tokens or 0) + (
                add_usage.reasoning_tokens or 0
            )
        else:
            reasoning_tokens = None

        if self.cached_tokens is not None or add_usage.cached_tokens is not None:
            cached_tokens = (self.cached_tokens or 0) + (add_usage.cached_tokens or 0)
        else:
            cached_tokens = None

        cost = (
            (self.cost or 0.0) + add_usage.cost
            if (add_usage.cost is not None)
            else None
        )
        return Usage(
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            reasoning_tokens=reasoning_tokens,
            cached_tokens=cached_tokens,
            cost=cost,
        )


class MessageBase(BaseModel):
    message_id: Hashable = Field(default_factory=lambda: str(uuid4())[:8])
    model_id: str | None = None


class AssistantMessage(MessageBase):
    role: Literal[Role.ASSISTANT] = Role.ASSISTANT
    content: str | None
    usage: Usage | None = None
    tool_calls: Sequence[ToolCall] | None = None
    refusal: str | None = None


class UserMessage(MessageBase):
    role: Literal[Role.USER] = Role.USER
    content: Content

    @classmethod
    def from_text(cls, text: str, model_id: str | None = None) -> "UserMessage":
        return cls(content=Content.from_text(text), model_id=model_id)

    @classmethod
    def from_formatted_prompt(
        cls,
        prompt_template: str,
        prompt_args: Mapping[str, str | int | bool | ImageData] | None = None,
        model_id: str | None = None,
    ) -> "UserMessage":
        content = Content.from_formatted_prompt(
            prompt_template=prompt_template, prompt_args=prompt_args
        )

        return cls(content=content, model_id=model_id)

    @classmethod
    def from_content_parts(
        cls,
        content_parts: Sequence[str | ImageData],
        model_id: str | None = None,
    ) -> "UserMessage":
        content = Content.from_content_parts(content_parts)

        return cls(content=content, model_id=model_id)


class SystemMessage(MessageBase):
    role: Literal[Role.SYSTEM] = Role.SYSTEM
    content: str


class ToolMessage(MessageBase):
    role: Literal[Role.TOOL] = Role.TOOL
    content: str
    tool_call_id: str

    @classmethod
    def from_tool_output(
        cls,
        tool_output: Any,
        tool_call: ToolCall,
        model_id: str | None = None,
        indent: int = 2,
    ) -> "ToolMessage":
        return cls(
            content=json.dumps(tool_output, default=pydantic_encoder, indent=indent),
            tool_call_id=tool_call.id,
            model_id=model_id,
        )


Message = Annotated[
    AssistantMessage | UserMessage | SystemMessage | ToolMessage,
    Field(discriminator="role"),
]

Conversation: TypeAlias = list[Message]



from collections.abc import Sequence
from pathlib import Path
from typing import Any, ClassVar, Generic, Protocol, cast, final

from pydantic import BaseModel

from .agent_message import AgentMessage
from .agent_message_pool import AgentMessagePool
from .comm_agent import CommunicatingAgent
from .llm import LLM, LLMSettings
from .llm_agent_state import (
    LLMAgentState,
    SetAgentState,
    SetAgentStateStrategy,
)
from .prompt_builder import (
    FormatInputArgsHandler,
    FormatSystemArgsHandler,
    PromptBuilder,
)
from .run_context import (
    CtxT,
    InteractionRecord,
    RunContextWrapper,
    SystemRunArgs,
    UserRunArgs,
)
from .tool_orchestrator import (
    ExitToolCallLoopHandler,
    ManageAgentStateHandler,
    ToolOrchestrator,
)
from .typing.content import ImageData
from .typing.converters import Converters
from .typing.io import (
    AgentID,
    AgentState,
    InT,
    LLMFormattedArgs,
    LLMFormattedSystemArgs,
    LLMPrompt,
    LLMPromptArgs,
    OutT,
)
from .typing.message import Conversation, Message, SystemMessage
from .typing.tool import BaseTool
from .utils import get_prompt, validate_obj_from_json_or_py_string


class ParseOutputHandler(Protocol[InT, OutT, CtxT]):
    def __call__(
        self,
        conversation: Conversation,
        *,
        in_args: InT | None,
        batch_idx: int,
        ctx: RunContextWrapper[CtxT] | None,
    ) -> OutT: ...


class LLMAgent(
    CommunicatingAgent[InT, OutT, LLMAgentState, CtxT],
    Generic[InT, OutT, CtxT],
):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_in_type",
        1: "_out_type",
    }

    def __init__(
        self,
        agent_id: AgentID,
        *,
        # LLM
        llm: LLM[LLMSettings, Converters],
        # Input prompt template (combines user and received arguments)
        in_prompt: LLMPrompt | None = None,
        in_prompt_path: str | Path | None = None,
        # System prompt template
        sys_prompt: LLMPrompt | None = None,
        sys_prompt_path: str | Path | None = None,
        # System args (static args provided via RunContextWrapper)
        sys_args_schema: type[LLMPromptArgs] = LLMPromptArgs,
        # User args (static args provided via RunContextWrapper)
        usr_args_schema: type[LLMPromptArgs] = LLMPromptArgs,
        # Tools
        tools: list[BaseTool[Any, Any, CtxT]] | None = None,
        max_turns: int = 1000,
        react_mode: bool = False,
        # Agent state management
        set_state_strategy: SetAgentStateStrategy = "keep",
        # Multi-agent routing
        message_pool: AgentMessagePool[CtxT] | None = None,
        recipient_ids: list[AgentID] | None = None,
    ) -> None:
        super().__init__(
            agent_id=agent_id, message_pool=message_pool, recipient_ids=recipient_ids
        )

        # Agent state
        self._state: LLMAgentState = LLMAgentState()
        self.set_state_strategy: SetAgentStateStrategy = set_state_strategy
        self._set_agent_state_impl: SetAgentState | None = None

        # Tool orchestrator

        self._using_default_llm_response_format: bool = False
        if llm.response_format is None and tools is None:
            llm.response_format = self.out_type
            self._using_default_llm_response_format = True

        self._tool_orchestrator: ToolOrchestrator[CtxT] = ToolOrchestrator[CtxT](
            agent_id=self.agent_id,
            llm=llm,
            tools=tools,
            max_turns=max_turns,
            react_mode=react_mode,
        )

        # Prompt builder
        sys_prompt = get_prompt(prompt_text=sys_prompt, prompt_path=sys_prompt_path)
        in_prompt = get_prompt(prompt_text=in_prompt, prompt_path=in_prompt_path)
        self._prompt_builder: PromptBuilder[InT, CtxT] = PromptBuilder[
            self.in_type, CtxT
        ](
            agent_id=self._agent_id,
            sys_prompt=sys_prompt,
            in_prompt=in_prompt,
            sys_args_schema=sys_args_schema,
            usr_args_schema=usr_args_schema,
        )

        self.no_tqdm = getattr(llm, "no_tqdm", False)

        self._register_overridden_handlers()

    @property
    def llm(self) -> LLM[LLMSettings, Converters]:
        return self._tool_orchestrator.llm

    @property
    def tools(self) -> dict[str, BaseTool[BaseModel, Any, CtxT]]:
        return self._tool_orchestrator.tools

    @property
    def max_turns(self) -> int:
        return self._tool_orchestrator.max_turns

    @property
    def sys_args_schema(self) -> type[LLMPromptArgs]:
        return self._prompt_builder.sys_args_schema

    @property
    def usr_args_schema(self) -> type[LLMPromptArgs]:
        return self._prompt_builder.usr_args_schema

    @property
    def sys_prompt(self) -> LLMPrompt | None:
        return self._prompt_builder.sys_prompt

    @property
    def in_prompt(self) -> LLMPrompt | None:
        return self._prompt_builder.in_prompt

    def _parse_output(
        self,
        conversation: Conversation,
        *,
        in_args: InT | None = None,
        batch_idx: int = 0,
        ctx: RunContextWrapper[CtxT] | None = None,
    ) -> OutT:
        if self._parse_output_impl:
            if self._using_default_llm_response_format:
                # When using custom output parsing, the required LLM response format
                # can differ from the final agent output type ->
                # set it back to None unless it was specified explicitly at init.
                self._tool_orchestrator.llm.response_format = None
                self._using_default_llm_response_format = False

            return self._parse_output_impl(
                conversation=conversation,
                in_args=in_args,
                batch_idx=batch_idx,
                ctx=ctx,
            )

        return validate_obj_from_json_or_py_string(
            str(conversation[-1].content or ""),
            adapter=self._out_type_adapter,
            from_substring=True,
        )

    @staticmethod
    def _validate_run_inputs(
        chat_inputs: LLMPrompt | Sequence[str | ImageData] | None = None,
        in_args: InT | Sequence[InT] | None = None,
        in_message: AgentMessage[InT, AgentState] | None = None,
        entry_point: bool = False,
    ) -> None:
        multiple_inputs_err_message = (
            "Only one of chat_inputs, in_args, or in_message must be provided."
        )
        if chat_inputs is not None and in_args is not None:
            raise ValueError(multiple_inputs_err_message)
        if chat_inputs is not None and in_message is not None:
            raise ValueError(multiple_inputs_err_message)
        if in_args is not None and in_message is not None:
            raise ValueError(multiple_inputs_err_message)

        if entry_point and in_message is not None:
            raise ValueError(
                "Entry point agent cannot receive messages from other agents."
            )

    @final
    async def run(
        self,
        chat_inputs: LLMPrompt | Sequence[str | ImageData] | None = None,
        *,
        in_message: AgentMessage[InT, AgentState] | None = None,
        in_args: InT | Sequence[InT] | None = None,
        entry_point: bool = False,
        ctx: RunContextWrapper[CtxT] | None = None,
        forbid_state_change: bool = False,
        **gen_kwargs: Any,  # noqa: ARG002
    ) -> AgentMessage[OutT, LLMAgentState]:
        # Get run arguments
        sys_args: SystemRunArgs = LLMPromptArgs()
        usr_args: UserRunArgs = LLMPromptArgs()
        if ctx is not None:
            run_args = ctx.run_args.get(self.agent_id)
            if run_args is not None:
                sys_args = run_args.sys
                usr_args = run_args.usr

        self._validate_run_inputs(
            chat_inputs=chat_inputs,
            in_args=in_args,
            in_message=in_message,
            entry_point=entry_point,
        )

        # 1. Make system prompt (can be None)
        formatted_sys_prompt = self._prompt_builder.make_sys_prompt(
            sys_args=sys_args, ctx=ctx
        )

        # 2. Set agent state

        cur_state = self.state.model_copy(deep=True)
        in_state = in_message.sender_state if in_message else None
        prev_mh_len = len(cur_state.message_history)

        state = LLMAgentState.from_cur_and_in_states(
            cur_state=cur_state,
            in_state=in_state,
            sys_prompt=formatted_sys_prompt,
            strategy=self.set_state_strategy,
            set_agent_state_impl=self._set_agent_state_impl,
            ctx=ctx,
        )

        self._print_sys_msg(state=state, prev_mh_len=prev_mh_len, ctx=ctx)

        # 3. Make and add user messages (can be empty)
        _in_args_batch: Sequence[InT] | None = None
        if in_message is not None:
            _in_args_batch = in_message.payloads
        elif in_args is not None:
            _in_args_batch = in_args if isinstance(in_args, Sequence) else [in_args]  # type: ignore[assignment]

        user_message_batch = self._prompt_builder.make_user_messages(
            chat_inputs=chat_inputs,
            usr_args=usr_args,
            in_args_batch=_in_args_batch,
            entry_point=entry_point,
            ctx=ctx,
        )
        if user_message_batch:
            state.message_history.add_message_batch(user_message_batch)
            self._print_msgs(user_message_batch, ctx=ctx)

        if not self.tools:
            # 4. Generate messages without tools
            await self._tool_orchestrator.generate_once(state=state, ctx=ctx)
        else:
            # 4. Run tool call loop (new messages are added to the message
            #    history inside the loop)
            await self._tool_orchestrator.run_loop(state=state, ctx=ctx)

        # 5. Parse outputs
        batch_size = state.message_history.batch_size
        in_args_batch = in_message.payloads if in_message else batch_size * [None]
        val_output_batch = [
            self._out_type_adapter.validate_python(
                self._parse_output(conversation=conv, in_args=in_args, ctx=ctx)
            )
            for conv, in_args in zip(
                state.message_history.batched_conversations,
                in_args_batch,
                strict=False,
            )
        ]

        # 6. Write interaction history to context

        recipient_ids = self._validate_routing(val_output_batch)

        if ctx:
            interaction_record = InteractionRecord(
                source_id=self.agent_id,
                recipient_ids=recipient_ids,
                chat_inputs=chat_inputs,
                sys_prompt=self.sys_prompt,
                in_prompt=self.in_prompt,
                sys_args=sys_args,
                usr_args=usr_args,
                in_args=(in_message.payloads if in_message is not None else None),
                outputs=val_output_batch,
                state=state,
            )
            ctx.interaction_history.append(
                cast(
                    "InteractionRecord[Any, Any, AgentState]",
                    interaction_record,
                )
            )

        agent_message = AgentMessage(
            payloads=val_output_batch,
            sender_id=self.agent_id,
            sender_state=state,
            recipient_ids=recipient_ids,
        )

        if not forbid_state_change:
            self._state = state

        return agent_message

    def _print_msgs(
        self,
        messages: Sequence[Message],
        ctx: RunContextWrapper[CtxT] | None = None,
    ) -> None:
        if ctx:
            ctx.printer.print_llm_messages(messages, agent_id=self.agent_id)

    def _print_sys_msg(
        self,
        state: LLMAgentState,
        prev_mh_len: int,
        ctx: RunContextWrapper[CtxT] | None = None,
    ) -> None:
        if (
            len(state.message_history) == 1
            and prev_mh_len == 0
            and isinstance(state.message_history[0][0], SystemMessage)
        ):
            self._print_msgs([state.message_history[0][0]], ctx=ctx)

    # -- Handlers for custom implementations --

    def format_sys_args_handler(
        self, func: FormatSystemArgsHandler[CtxT]
    ) -> FormatSystemArgsHandler[CtxT]:
        self._prompt_builder.format_sys_args_impl = func

        return func

    def format_in_args_handler(
        self, func: FormatInputArgsHandler[InT, CtxT]
    ) -> FormatInputArgsHandler[InT, CtxT]:
        self._prompt_builder.format_in_args_impl = func

        return func

    def parse_output_handler(
        self, func: ParseOutputHandler[InT, OutT, CtxT]
    ) -> ParseOutputHandler[InT, OutT, CtxT]:
        self._parse_output_impl = func

        return func

    def set_agent_state_handler(self, func: SetAgentState) -> SetAgentState:
        self._make_custom_agent_state_impl = func

        return func

    def exit_tool_call_loop_handler(
        self, func: ExitToolCallLoopHandler[CtxT]
    ) -> ExitToolCallLoopHandler[CtxT]:
        self._tool_orchestrator.exit_tool_call_loop_impl = func

        return func

    def manage_agent_state_handler(
        self, func: ManageAgentStateHandler[CtxT]
    ) -> ManageAgentStateHandler[CtxT]:
        self._tool_orchestrator.manage_agent_state_impl = func

        return func

    # -- Override these methods in subclasses if needed --

    def _register_overridden_handlers(self) -> None:
        cur_cls = type(self)
        base_cls = LLMAgent[Any, Any, Any]

        if cur_cls._format_sys_args is not base_cls._format_sys_args:  # noqa: SLF001
            self._prompt_builder.format_sys_args_impl = self._format_sys_args

        if cur_cls._format_in_args is not base_cls._format_in_args:  # noqa: SLF001
            self._prompt_builder.format_in_args_impl = self._format_in_args

        if cur_cls._set_agent_state is not base_cls._set_agent_state:  # noqa: SLF001
            self._set_agent_state_impl = self._set_agent_state

        if cur_cls._manage_agent_state is not base_cls._manage_agent_state:  # noqa: SLF001
            self._tool_orchestrator.manage_agent_state_impl = self._manage_agent_state

        if (
            cur_cls._exit_tool_call_loop is not base_cls._exit_tool_call_loop  # noqa: SLF001
        ):
            self._tool_orchestrator.exit_tool_call_loop_impl = self._exit_tool_call_loop

        self._parse_output_impl: ParseOutputHandler[InT, OutT, CtxT] | None = None

    def _format_sys_args(
        self,
        sys_args: LLMPromptArgs,
        *,
        ctx: RunContextWrapper[CtxT] | None = None,
    ) -> LLMFormattedSystemArgs:
        raise NotImplementedError(
            "LLMAgent._format_sys_args must be overridden by a subclass "
            "if it's intended to be used as the system arguments formatter."
        )

    def _format_in_args(
        self,
        *,
        usr_args: LLMPromptArgs,
        in_args: InT,
        batch_idx: int = 0,
        ctx: RunContextWrapper[CtxT] | None = None,
    ) -> LLMFormattedArgs:
        raise NotImplementedError(
            "LLMAgent._format_in_args must be overridden by a subclass"
        )

    def _set_agent_state(
        self,
        cur_state: LLMAgentState,
        *,
        in_state: AgentState | None,
        sys_prompt: LLMPrompt | None,
        ctx: RunContextWrapper[Any] | None,
    ) -> LLMAgentState:
        raise NotImplementedError(
            "LLMAgent._set_agent_state_handler must be overridden by a subclass"
        )

    def _exit_tool_call_loop(
        self,
        conversation: Conversation,
        *,
        ctx: RunContextWrapper[CtxT] | None = None,
        **kwargs: Any,
    ) -> bool:
        raise NotImplementedError(
            "LLMAgent._tool_call_loop_exit must be overridden by a subclass"
        )

    def _manage_agent_state(
        self,
        state: LLMAgentState,
        *,
        ctx: RunContextWrapper[CtxT] | None = None,
        **kwargs: Any,
    ) -> None:
        raise NotImplementedError(
            "LLMAgent._manage_agent_state must be overridden by a subclass"
        )



from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Any, ClassVar, Generic

from ..agent_message_pool import AgentMessage, AgentMessagePool
from ..comm_agent import CommunicatingAgent
from ..run_context import CtxT, RunContextWrapper
from ..typing.io import AgentID, InT, OutT


class WorkflowAgent(
    CommunicatingAgent[InT, OutT, Any, CtxT], ABC, Generic[InT, OutT, CtxT]
):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_in_type",
        1: "_out_type",
    }

    def __init__(
        self,
        agent_id: AgentID,
        subagents: Sequence[CommunicatingAgent[Any, Any, Any, CtxT]],
        start_agent: CommunicatingAgent[InT, Any, Any, CtxT],
        end_agent: CommunicatingAgent[Any, OutT, Any, CtxT],
        message_pool: AgentMessagePool[CtxT] | None = None,
        recipient_ids: list[AgentID] | None = None,
        dynamic_routing: bool = False,
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        if not subagents:
            raise ValueError("At least one step is required")
        if start_agent not in subagents:
            raise ValueError("Start agent must be in the subagents list")
        if end_agent not in subagents:
            raise ValueError("End agent must be in the subagents list")

        self.subagents = subagents

        self._start_agent = start_agent
        self._end_agent = end_agent

        super().__init__(
            agent_id=agent_id,
            message_pool=message_pool,
            recipient_ids=recipient_ids,
            dynamic_routing=dynamic_routing,
        )
        for subagent in subagents:
            assert not subagent.recipient_ids, (
                "Subagents must not have recipient_ids set."
            )

    @property
    def start_agent(self) -> CommunicatingAgent[InT, Any, Any, CtxT]:
        return self._start_agent

    @property
    def end_agent(self) -> CommunicatingAgent[Any, OutT, Any, CtxT]:
        return self._end_agent

    @abstractmethod
    async def run(
        self,
        chat_inputs: Any | None = None,
        *,
        in_args: InT | Sequence[InT] | None = None,
        in_message: AgentMessage[InT, Any] | None = None,
        ctx: RunContextWrapper[CtxT] | None = None,
        entry_point: bool = False,
        forbid_state_change: bool = False,
        **generation_kwargs: Any,
    ) -> AgentMessage[OutT, Any]:
        pass



import logging
from abc import abstractmethod
from collections.abc import Sequence
from typing import Any, ClassVar, Generic, Protocol, TypeVar, cast, final

from pydantic import BaseModel, TypeAdapter
from pydantic.json_schema import SkipJsonSchema

from .agent_message import AgentMessage
from .agent_message_pool import AgentMessagePool
from .base_agent import BaseAgent
from .run_context import CtxT, RunContextWrapper
from .typing.io import AgentID, AgentState, InT, OutT, StateT
from .typing.tool import BaseTool

logger = logging.getLogger(__name__)


class DynCommPayload(BaseModel):
    selected_recipient_ids: SkipJsonSchema[Sequence[AgentID]]


_EH_OutT = TypeVar("_EH_OutT", contravariant=True)  # noqa: PLC0105
_EH_StateT = TypeVar("_EH_StateT", bound=AgentState, contravariant=True)  # noqa: PLC0105


class ExitHandler(Protocol[_EH_OutT, _EH_StateT, CtxT]):
    def __call__(
        self,
        output_message: AgentMessage[_EH_OutT, _EH_StateT],
        ctx: RunContextWrapper[CtxT] | None,
    ) -> bool: ...


class CommunicatingAgent(
    BaseAgent[OutT, StateT, CtxT], Generic[InT, OutT, StateT, CtxT]
):
    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_in_type",
        1: "_out_type",
    }

    def __init__(
        self,
        agent_id: AgentID,
        *,
        recipient_ids: Sequence[AgentID] | None = None,
        message_pool: AgentMessagePool[CtxT] | None = None,
        **kwargs: Any,
    ) -> None:
        self._in_type: type[InT]
        super().__init__(agent_id=agent_id, **kwargs)

        self._in_args_type_adapter: TypeAdapter[InT] = TypeAdapter(self._in_type)
        self.recipient_ids = recipient_ids or []

        self._message_pool = message_pool or AgentMessagePool()
        self._is_listening = False
        self._exit_impl: ExitHandler[OutT, StateT, CtxT] | None = None

    @property
    def in_type(self) -> type[InT]:  # type: ignore
        # Exposing the type of a contravariant variable only, should be safe
        return self._in_type

    def _validate_routing(self, payloads: Sequence[OutT]) -> Sequence[AgentID]:
        if all(isinstance(p, DynCommPayload) for p in payloads):
            payloads_ = cast("Sequence[DynCommPayload]", payloads)
            selected_recipient_ids_per_payload = [
                set(p.selected_recipient_ids or []) for p in payloads_
            ]
            assert all(
                x == selected_recipient_ids_per_payload[0]
                for x in selected_recipient_ids_per_payload
            ), "All payloads must have the same recipient IDs for dynamic routing"

            assert payloads_[0].selected_recipient_ids is not None
            selected_recipient_ids = payloads_[0].selected_recipient_ids

            assert all(rid in self.recipient_ids for rid in selected_recipient_ids), (
                "Dynamic routing is enabled, but recipient IDs are not in "
                "the allowed agent's recipient IDs"
            )

            return selected_recipient_ids

        if all((not isinstance(p, DynCommPayload)) for p in payloads):
            return self.recipient_ids

        raise ValueError(
            "All payloads must be either DCommAgentPayload or not DCommAgentPayload"
        )

    async def post_message(self, message: AgentMessage[OutT, StateT]) -> None:
        self._validate_routing(message.payloads)

        await self._message_pool.post(message)

    @abstractmethod
    async def run(
        self,
        chat_inputs: Any | None = None,
        *,
        ctx: RunContextWrapper[CtxT] | None = None,
        in_message: AgentMessage[InT, AgentState] | None = None,
        entry_point: bool = False,
        forbid_state_change: bool = False,
        **kwargs: Any,
    ) -> AgentMessage[OutT, StateT]:
        pass

    async def run_and_post(
        self, ctx: RunContextWrapper[CtxT] | None = None, **run_kwargs: Any
    ) -> None:
        output_message = await self.run(
            ctx=ctx, in_message=None, entry_point=True, **run_kwargs
        )
        await self.post_message(output_message)

    def exit_handler(
        self, func: ExitHandler[OutT, StateT, CtxT]
    ) -> ExitHandler[OutT, StateT, CtxT]:
        self._exit_impl = func

        return func

    def _exit_condition(
        self,
        output_message: AgentMessage[OutT, StateT],
        ctx: RunContextWrapper[CtxT] | None,
    ) -> bool:
        if self._exit_impl:
            return self._exit_impl(output_message=output_message, ctx=ctx)

        return False

    async def _message_handler(
        self,
        message: AgentMessage[Any, AgentState],
        ctx: RunContextWrapper[CtxT] | None = None,
        **run_kwargs: Any,
    ) -> None:
        in_message = cast("AgentMessage[InT, AgentState]", message)
        out_message = await self.run(ctx=ctx, in_message=in_message, **run_kwargs)

        if self._exit_condition(output_message=out_message, ctx=ctx):
            await self._message_pool.stop_all()
            return

        if self.recipient_ids:
            await self.post_message(out_message)

    @property
    def is_listening(self) -> bool:
        return self._is_listening

    async def start_listening(
        self, ctx: RunContextWrapper[CtxT] | None = None, **run_kwargs: Any
    ) -> None:
        if self._is_listening:
            return

        self._is_listening = True
        self._message_pool.register_message_handler(
            agent_id=self.agent_id,
            handler=self._message_handler,
            ctx=ctx,
            **run_kwargs,
        )

    async def stop_listening(self) -> None:
        self._is_listening = False
        await self._message_pool.unregister_message_handler(self.agent_id)

    @final
    def as_tool(
        self,
        tool_name: str,
        tool_description: str,
        tool_strict: bool = True,
    ) -> BaseTool[InT, OutT, Any]:  # type: ignore[override]
        # Will check if InT is a BaseModel at runtime
        agent_instance = self
        in_type = agent_instance.in_type
        out_type = agent_instance.out_type
        if not issubclass(in_type, BaseModel):
            raise TypeError(
                "Cannot create a tool from an agent with "
                f"non-BaseModel input type: {in_type}"
            )

        class AgentTool(BaseTool[in_type, out_type, Any]):
            name: str = tool_name
            description: str = tool_description
            strict: bool | None = tool_strict

            async def run(
                self,
                inp: InT,
                ctx: RunContextWrapper[CtxT] | None = None,
            ) -> OutT:
                in_args = in_type.model_validate(inp)
                in_message = AgentMessage[in_type, AgentState](
                    payloads=[in_args],
                    sender_id="<tool_user>",
                    recipient_ids=[agent_instance.agent_id],
                )
                agent_result = await agent_instance.run(
                    in_message=in_message,
                    entry_point=False,
                    forbid_state_change=True,
                    ctx=ctx,
                )

                return agent_result.payloads[0]

        return AgentTool()  # type: ignore[return-value]



from abc import ABC

from pydantic import BaseModel

from .message import AssistantMessage


class CompletionChoice(BaseModel):
    # TODO: add fields
    message: AssistantMessage
    finish_reason: str | None


class CompletionError(BaseModel):
    message: str
    metadata: dict[str, str | None] | None = None
    code: int


class Completion(BaseModel, ABC):
    # TODO: add fields
    choices: list[CompletionChoice]
    model_id: str | None = None
    error: CompletionError | None = None


class CompletionChunk(BaseModel):
    # TODO: add more fields and tool use support (and choices?)
    delta: str | None = None
    model_id: str | None = None
